@@ -2,7 +2,7 @@
 
 # ibus-table - The Tables engine for IBus
 #
-# Copyright (c) 2018 Mike FABIAN <mfabian@redhat.com>
+# Copyright (c) 2018, 2021 Mike FABIAN <mfabian@redhat.com>
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -29,7 +29,7 @@ import logging
 import time
 import unittest
 import importlib
-import mock
+from unittest import mock
 
 from gi import require_version
 require_version('IBus', '1.0')
@@ -86,6 +86,7 @@ ORIG_MULTI_WILDCARD_CHAR = None
 ORIG_PINYIN_MODE = None
 ORIG_SUGGESTION_MODE = None
 ORIG_KEYBINDINGS = None
+ORIG_DYNAMIC_ADJUST = None
 ORIG_DEBUG_LEVEL = None
 
 def backup_original_settings():
@@ -106,6 +107,7 @@ def backup_original_settings():
     global ORIG_MULTI_WILDCARD_CHAR
     global ORIG_PINYIN_MODE
     global ORIG_SUGGESTION_MODE
+    global ORIG_DYNAMIC_ADJUST
     global ORIG_DEBUG_LEVEL
     ORIG_KEYBINDINGS = ENGINE.get_keybindings()
     ORIG_INPUT_MODE = ENGINE.get_input_mode()
@@ -123,6 +125,7 @@ def backup_original_settings():
     ORIG_MULTI_WILDCARD_CHAR = ENGINE.get_multi_wildcard_char()
     ORIG_PINYIN_MODE = ENGINE.get_pinyin_mode()
     ORIG_SUGGESTION_MODE = ENGINE.get_suggestion_mode()
+    ORIG_DYNAMIC_ADJUST = ENGINE.get_dynamic_adjust()
     ORIG_DEBUG_LEVEL = ENGINE.get_debug_level()
 
 def restore_original_settings():
@@ -143,6 +146,7 @@ def restore_original_settings():
     global ORIG_MULTI_WILDCARD_CHAR
     global ORIG_PINYIN_MODE
     global ORIG_SUGGESTION_MODE
+    global ORIG_DYNAMIC_ADJUST
     global ORIG_DEBUG_LEVEL
     ENGINE.set_keybindings(
         ORIG_KEYBINDINGS, update_gsettings=False)
@@ -176,6 +180,7 @@ def restore_original_settings():
         ORIG_MULTI_WILDCARD_CHAR, update_gsettings=False)
     ENGINE.set_pinyin_mode(ORIG_PINYIN_MODE)
     ENGINE.set_suggestion_mode(ORIG_SUGGESTION_MODE)
+    ENGINE.set_dynamic_adjust(ORIG_DYNAMIC_ADJUST)
     ENGINE.set_debug_level(ORIG_DEBUG_LEVEL, update_gsettings=False)
 
 def set_default_settings():
@@ -233,6 +238,15 @@ def set_default_settings():
     ENGINE.set_onechar_mode(
         onechar, update_gsettings=False)
 
+    dynamic_adjust = False
+    dynamic_adjust = TABSQLITEDB.ime_properties.get('dynamic_adjust')
+    if dynamic_adjust:
+        dynamic_adjust = (dynamic_adjust.lower() == u'true')
+    else:
+        dynamic_adjust = True
+    ENGINE.set_dynamic_adjust(
+        dynamic_adjust, update_gsettings=False)
+
     auto_select_mode = False
     auto_select = TABSQLITEDB.ime_properties.get('auto_select')
     if auto_select:
@@ -720,9 +734,9 @@ class WubiJidian86TestCase(unittest.TestCase):
                           '蚂    jcg 3110000 0',
                           '杩    scg 1280000 0',
                           '犸    qtcg 120000 0',
-                          '溤    icy 38700 0',
                           '鰢    qocy 36500 0',
-                          '鷌    wvgc 25500 0'])
+                          '鷌    wvgc 25500 0',
+                          '㐷    wcg 9070 0'])
         ENGINE._do_process_key_event(IBus.KEY_space, 0, 0)
         self.assertEqual(ENGINE.mock_preedit_text, '')
         self.assertEqual(ENGINE.mock_committed_text, '吗')
@@ -758,9 +772,9 @@ class WubiJidian86TestCase(unittest.TestCase):
                           '蚂    jcg 3110000 0',
                           '杩    scg 1280000 0',
                           '犸    qtcg 120000 0',
-                          '溤    icy 38700 0',
                           '鰢    qocy 36500 0',
-                          '鷌    wvgc 25500 0'])
+                          '鷌    wvgc 25500 0',
+                          '㐷    wcg 9070 0'])
         ENGINE._do_process_key_event(IBus.KEY_space, 0, 0)
         self.assertEqual(ENGINE.mock_preedit_text, '')
         self.assertEqual(ENGINE.mock_committed_text, '吗嗎吗')
@@ -876,37 +890,41 @@ class WubiJidian86TestCase(unittest.TestCase):
         # the left or the right shift key:
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_L, 0,
-            IBus.ModifierType.SHIFT_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_L, 0,
-            IBus.ModifierType.SHIFT_MASK | IBus.ModifierType.RELEASE_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
         self.assertEqual(ENGINE.mock_preedit_text, '工')
         self.assertEqual(ENGINE.mock_committed_text, '')
         ENGINE._do_process_key_event(IBus.KEY_b, 0, 0)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_R, 0,
-            IBus.ModifierType.SHIFT_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_R, 0,
-            IBus.ModifierType.SHIFT_MASK | IBus.ModifierType.RELEASE_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
         self.assertEqual(ENGINE.mock_preedit_text, '工了')
         self.assertEqual(ENGINE.mock_committed_text, '')
         ENGINE._do_process_key_event(IBus.KEY_c, 0, 0)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_R, 0,
-            IBus.ModifierType.SHIFT_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_R, 0,
-            IBus.ModifierType.SHIFT_MASK | IBus.ModifierType.RELEASE_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
         self.assertEqual(ENGINE.mock_preedit_text, '工了以')
         self.assertEqual(ENGINE.mock_committed_text, '')
         ENGINE._do_process_key_event(IBus.KEY_d, 0, 0)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_L, 0,
-            IBus.ModifierType.SHIFT_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_L, 0,
-            IBus.ModifierType.SHIFT_MASK | IBus.ModifierType.RELEASE_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
         self.assertEqual(ENGINE.mock_preedit_text, '工了以在')
         self.assertEqual(ENGINE.mock_committed_text, '')
         # Move left two characters in the preëdit:
@@ -934,10 +952,11 @@ class WubiJidian86TestCase(unittest.TestCase):
         self.assertEqual(ENGINE.mock_preedit_text, '工了你以在')
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_L, 0,
-            IBus.ModifierType.SHIFT_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_L, 0,
-            IBus.ModifierType.SHIFT_MASK | IBus.ModifierType.RELEASE_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
         self.assertEqual(ENGINE.mock_preedit_text, '工了你以在')
         self.assertEqual(ENGINE.mock_committed_text, '')
         ENGINE._do_process_key_event(IBus.KEY_h, 0, 0)
@@ -950,10 +969,11 @@ class WubiJidian86TestCase(unittest.TestCase):
         self.assertEqual(ENGINE.mock_committed_text, '')
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_L, 0,
-            IBus.ModifierType.SHIFT_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK)
         ENGINE._do_process_key_event(
             IBus.KEY_Shift_L, 0,
-            IBus.ModifierType.SHIFT_MASK | IBus.ModifierType.RELEASE_MASK)
+            IBus.ModifierType.CONTROL_MASK | IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
         self.assertEqual(ENGINE.mock_preedit_text, '工了你好以在')
         self.assertEqual(ENGINE.mock_committed_text, '')
         # Move right two characters in the preëdit
@@ -990,6 +1010,61 @@ class WubiJidian86TestCase(unittest.TestCase):
         ENGINE._do_process_key_event(IBus.KEY_space, 0, 0)
         self.assertEqual(ENGINE.mock_committed_text, '工了你好以在工了你好以在')
 
+    def test_switch_to_direct_mode_and_commit_english(self):
+        '''
+        When in Chinese mode and the preedit is not empty, switching to
+        direct mode should be possible and it should not discard the
+        input but commit as English.
+
+        See: https://github.com/kaio/ibus-table/issues/68
+        '''
+        ENGINE._do_process_key_event(IBus.KEY_a, 0, 0)
+        self.assertEqual(ENGINE.mock_preedit_text, '工')
+        self.assertEqual(ENGINE.mock_committed_text, '')
+        # Shift_L is the default key binding to switch to direct mode
+        ENGINE._do_process_key_event(
+            IBus.KEY_Shift_L, 0,
+            IBus.ModifierType.SHIFT_MASK)
+        ENGINE._do_process_key_event(
+            IBus.KEY_Shift_L, 0,
+            IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
+        self.assertEqual(ENGINE.mock_preedit_text, '')
+        self.assertEqual(ENGINE.mock_committed_text, 'a')
+
+    def test_switch_between_table_and_pinyin_mode(self):
+        '''
+        The switch between table and pinyin mode should happen immediately
+        even if the preedit is not empty
+        '''
+        ENGINE._do_process_key_event(IBus.KEY_a, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
+        self.assertEqual(ENGINE.mock_preedit_text, '东')
+        self.assertEqual(ENGINE.mock_committed_text, '')
+        # Shift_R is the default key binding to switch between table
+        # and pinyin mode
+        ENGINE._do_process_key_event(
+            IBus.KEY_Shift_R, 0,
+            IBus.ModifierType.SHIFT_MASK)
+        ENGINE._do_process_key_event(
+            IBus.KEY_Shift_R, 0,
+            IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
+        self.assertEqual(ENGINE.mock_preedit_text, '爱')
+        self.assertEqual(ENGINE.mock_committed_text, '')
+        ENGINE._do_process_key_event(
+            IBus.KEY_Shift_R, 0,
+            IBus.ModifierType.SHIFT_MASK)
+        ENGINE._do_process_key_event(
+            IBus.KEY_Shift_R, 0,
+            IBus.ModifierType.SHIFT_MASK
+            | IBus.ModifierType.RELEASE_MASK)
+        self.assertEqual(ENGINE.mock_preedit_text, '东')
+        self.assertEqual(ENGINE.mock_committed_text, '')
+        ENGINE._do_process_key_event(IBus.KEY_space, 0, 0)
+        self.assertEqual(ENGINE.mock_preedit_text, '')
+        self.assertEqual(ENGINE.mock_committed_text, '东')
+
     def test_chinese_mode(self):
         ENGINE.set_chinese_mode(
             mode=0, update_gsettings=False) # show simplified Chinese only
@@ -1216,6 +1291,136 @@ class Cangjie5TestCase(unittest.TestCase):
         ENGINE._do_process_key_event(IBus.KEY_space, 0, 0)
         self.assertEqual(ENGINE.mock_committed_text, '機')
 
+    def test_dynamic_adjust(self):
+        ENGINE.set_dynamic_adjust(False)
+        ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
+        self.assertEqual('怠', ENGINE.mock_preedit_text)
+        # Big5 code of 怠 is ABE5 (see
+        # https://www.unicode.org/cgi-bin/GetUnihanData.pl?codepoint=%E6%84%9F)
+        # Big5 code of 感 is B750 (see
+        # https://www.unicode.org/cgi-bin/GetUnihanData.pl?codepoint=%E6%80%A0)
+        self.assertEqual(
+            ['怠  1000 0',
+             '感  1000 0'],
+            ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_2, 0, 0)
+        self.assertEqual(
+            [],
+            ENGINE._lookup_table.mock_candidates)
+        self.assertEqual('', ENGINE.mock_preedit_text)
+        self.assertEqual('感', ENGINE.mock_committed_text)
+        # repeat the same input, result should be the same
+        # because dynamic adjust is False:
+        ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
+        self.assertEqual('怠', ENGINE.mock_preedit_text)
+        self.assertEqual(
+            ['怠  1000 0',
+             '感  1000 0'],
+            ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_2, 0, 0)
+        self.assertEqual(
+            [],
+            ENGINE._lookup_table.mock_candidates)
+        self.assertEqual('', ENGINE.mock_preedit_text)
+        self.assertEqual('感感', ENGINE.mock_committed_text)
+        # Now set dynamic adjust to True:
+        ENGINE.set_dynamic_adjust(True)
+        # repeat the same input, result should still be the same
+        # because no dynamic adjustment has happened yet:
+        ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
+        self.assertEqual('怠', ENGINE.mock_preedit_text)
+        self.assertEqual(
+            ['怠  1000 0',
+             '感  1000 0'],
+            ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_2, 0, 0)
+        self.assertEqual(
+            [],
+            ENGINE._lookup_table.mock_candidates)
+        self.assertEqual('', ENGINE.mock_preedit_text)
+        self.assertEqual('感感感', ENGINE.mock_committed_text)
+        # Try once again, now 感 should be preferred because
+        # 感 has been selected once while dynamic adjustment was used
+        # whereas 怠 has not been selected:
+        ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
+        self.assertEqual('感', ENGINE.mock_preedit_text)
+        self.assertEqual(
+            ['感  1000 1',
+             '怠  1000 0'],
+            ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_1, 0, 0)
+        self.assertEqual(
+            [],
+            ENGINE._lookup_table.mock_candidates)
+        self.assertEqual('', ENGINE.mock_preedit_text)
+        self.assertEqual('感感感感', ENGINE.mock_committed_text)
+        # Try once again with dynamic adjust, now 感 should have
+        # user count 2:
+        ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
+        self.assertEqual('感', ENGINE.mock_preedit_text)
+        self.assertEqual(
+            ['感  1000 2',
+             '怠  1000 0'],
+            ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_1, 0, 0)
+        self.assertEqual(
+            [],
+            ENGINE._lookup_table.mock_candidates)
+        self.assertEqual('', ENGINE.mock_preedit_text)
+        self.assertEqual('感感感感感', ENGINE.mock_committed_text)
+        # 感 should have user count 3 now. This should be remembered
+        # in the user data base, i.e. if we switch off dynamic adjust
+        # and later switch it back on, the previous user count should
+        # still be there.
+        #
+        # Switch dynamic adjust off again:
+        ENGINE.set_dynamic_adjust(False)
+        # And try again, now 怠 should be preferred again
+        # because of its smaller Big5 code and the user counts
+        # displayed should be 0:
+        ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
+        self.assertEqual('怠', ENGINE.mock_preedit_text)
+        self.assertEqual(
+            ['怠  1000 0',
+             '感  1000 0'],
+            ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_2, 0, 0)
+        self.assertEqual(
+            [],
+            ENGINE._lookup_table.mock_candidates)
+        self.assertEqual('', ENGINE.mock_preedit_text)
+        self.assertEqual('感感感感感感', ENGINE.mock_committed_text)
+        # Switch dynamic adjust on again:
+        ENGINE.set_dynamic_adjust(True)
+        # Try once again with dynamic adjust, now 感 should be preferred
+        # again and have user count 3:
+        ENGINE._do_process_key_event(IBus.KEY_i, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_r, 0, 0)
+        ENGINE._do_process_key_event(IBus.KEY_p, 0, 0)
+        self.assertEqual('感', ENGINE.mock_preedit_text)
+        self.assertEqual(
+            ['感  1000 3',
+             '怠  1000 0'],
+            ENGINE._lookup_table.mock_candidates)
+        ENGINE._do_process_key_event(IBus.KEY_1, 0, 0)
+        self.assertEqual(
+            [],
+            ENGINE._lookup_table.mock_candidates)
+        self.assertEqual('', ENGINE.mock_preedit_text)
+        self.assertEqual('感感感感感感感', ENGINE.mock_committed_text)
+
 class IpaXSampaTestCase(unittest.TestCase):
     def setUp(self):
         engine_name = 'ipa-x-sampa'
@@ -1308,6 +1513,12 @@ class LatexTestCase(unittest.TestCase):
         self.assertEqual(ENGINE.mock_committed_text, 'α\\alpha ')
 
     def test_single_char_commit_with_f3(self):
+        # The latex.txt table in ibus-table-others-1.3.12 has
+        # SELECT_KEYS = F1,F2,F3,F4,F5,F6,F7,F8,F9,F10
+        # the older version in ibus-table-others-1.3.11
+        # has SELECT_KEYS = F1,F2,F3,F4,F5,F6,F7,F8,F9
+        # To make this test work for both versions, set the page size to 9:
+        ENGINE.set_page_size(9)
         ENGINE._do_process_key_event(IBus.KEY_backslash, 0, 0)
         ENGINE._do_process_key_event(IBus.KEY_b, 0, 0)
         # Lookup table shows only the first page, subsequent
